from combinatorics import xuniqueCombinations 

class node:
    def __init__(self, id):
        self.id = id
        self.counter = 0
        self.in_ = set()
        self.out = set()
        
    def __str__(self):
        return str(self.__dict__)

def key( items ):
    return "->".join([x.id for x in items])

class pairs_storage:
    def __init__( self, n ):
        self.__n = n
        self.__nodes = {}
        self.__combs_arr = []
        for i in range(n): 
            self.__combs_arr.append( set() )
    
    def add( self, comb ):
        n = len(comb)
        assert(n>0)
        
        self.__combs_arr[n-1].add(key(comb))
        if n == 1 and comb[0].id not in self.__nodes:
            self.__nodes[comb[0].id] = node(comb[0].id)
            return
        
        ids = [x.id for x in comb]
        for i, id in enumerate(ids):
            curr = self.__nodes[id]
            curr.counter += 1
            curr.in_.update( ids[:i] )
            curr.out.update(ids[i+1:])

    def add_sequence( self, seq ):
        for i in range(1, self.__n+1):
            for comb in xuniqueCombinations(seq, i):
                self.add(comb)
    
    def get_node_info( self, item ):
        return self.__nodes.get( item.id, node(item.id) )
    
    def get_combs( self ):
        return self.__combs_arr
 
    def __len__( self ):
        return len(self.__combs_arr[-1])

    def count_new_combs( self, seq ):
        s = set([key(z) for z in xuniqueCombinations( seq, self.__n)]) - self.__combs_arr[-1]
        return len(s)



import functools

import metacomm.combinatorics.all_pairs2
all_pairs = metacomm.combinatorics.all_pairs2.all_pairs2


"""
Another demo of filtering capabilities.
Demonstrates how to use named parameters
"""

# sample parameters are is taken from 
# http://www.stsc.hill.af.mil/consulting/sw_testing/improvement/cst.html


parameters = [ ( "brand"
               , [ "Brand X", "Brand Y" ] )
             , ( "os"
               , [ "98", "NT", "2000", "XP"] )
             , ( "network"
               , [ "Internal", "Modem" ] )
             , ( "employee"
               , [ "Salaried", "Hourly", "Part-Time", "Contr." ] )
             , ( "increment"
               , [ 6, 10, 15, 30, 60 ] )
             ]

                 
def is_valid_combination( values, names ):

    dictionary = dict( zip( names, values ) )

    """
    Should return True if combination is valid and False otherwise.
    
    Dictionary that is passed here can be incomplete.
    To prevent search for unnecessary items filtering function
    is executed with found subset of data to validate it.
    """

    rules = [ lambda d: "98" == d["os"] and "Brand Y" == d["brand"] # Brand Y does not support Windows 98
            , lambda d: "XP" == d["os"] and "Brand X" == d["brand"] # Brand X does not work with XP
            , lambda d: "Contr." == d["employee"] and d["increment"] < 30 # Contractors are billed in 30 min increments
            ]
    
    for rule in rules:
        try:
            if rule(dictionary):
                return False
        except KeyError: pass
    
    return True
          
             
pairwise = all_pairs(
      [ x[1] for x in parameters ]
    , filter_func = lambda values: is_valid_combination( values, [ x[0] for x in parameters ] )
    )

for i, v in enumerate(pairwise):
    print "%i:\t%s" % (i, str(v))





        


import metacomm.combinatorics.all_pairs2
all_pairs = metacomm.combinatorics.all_pairs2.all_pairs2


"""
Provided to make it easier to compare efficiency with other tools
as per http://pairwise.org/tools.asp

Current iutput is:

3^4: produces 9 rows
3^13: produces 17 rows
4^15 * 3^17 * 2^29: produces 37 rows
4^1 * 3^39 * 2^35: produces 27 rows
3^100: produces 29 rows
10^20: produces 219 rows
10^10: produces 172 rows

"""

def get_arrays( dimensions ):
    opts = []
   
    for d in dimensions:
        r = []
        for i in range(d[1]):
            r.append( range(d[0]) )
        opts += r
            
    return opts

def print_result( dimensions ):
    header_list = []
    for d in dimensions:
        header_list.append( "%i^%i" % d )
    header = " * ".join(header_list)
    
    pairwise = all_pairs( get_arrays( dimensions ) )
    n = len(list(pairwise))
    
    print "%s: produces %i rows" % (header, n)
 
        
print_result(( (3, 4), ))
print_result(( (3, 13), ))
print_result(( (4, 15), (3, 17), (2, 29) ))
print_result(( (4, 1), (3, 39), (2, 35) ))
print_result(( (3, 100), ))
print_result(( (10, 20), ))
print_result(( (10, 10), ))





#===============================================================================
# code from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/190465
# all (c) etc. are of the authors of this procedures, see link above
#===============================================================================


def xcombinations(items, n):
    if n==0: yield []
    else:
        for i in xrange(len(items)):
            for cc in xcombinations(items[:i]+items[i+1:],n-1):
                yield [items[i]]+cc

def xuniqueCombinations(items, n):
    if n==0: yield []
    else:
        for i in xrange(len(items)):
            for cc in xuniqueCombinations( items[i+1:], n-1) :
                yield [items[i]]+cc
                            
def xselections(items, n):
    if n==0: yield []
    else:
        for i in xrange(len(items)):
            for ss in xselections(items, n-1):
                yield [items[i]]+ss

def xpermutations(items):
    return xcombinations(items, len(items))

#supposedly it is faster
def permutations2(L):
    if len(L) <= 1:
        yield L
    else:
        a = [L.pop(0)]
        for p in permutations(L):
            for i in range(len(p)+1):
                yield p[:i] + a + p[i:]




from distutils.core import setup

setup(
      name = 'AllPairs'
    , version = '2.0.1'
    , description = 'Pairwise test combinations generator'
    , long_description = '''Pairwise (aka "all-pairs") test combinations generator written in
Python. Allows one to create a set of tests using "pairwise 
combinations" method, reducing a number of combinations of variables
into a lesser set that covers most situations.
'''
    , author = 'MetaCommunications Engineering'
    , author_email = 'metacomm@users.sourceforge.net'
    , url='http://apps.sourceforge.net/trac/allpairs/'
    , packages = [ 'metacomm', 'metacomm.combinatorics' ]
    , classifiers =
        [
          'Development Status :: 5 - Production/Stable',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Information Technology',
          'Intended Audience :: Science/Research',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: MIT License',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.3',
          'Programming Language :: Python :: 2.4',
          'Programming Language :: Python :: 2.5',
          'Topic :: Software Development :: Testing',
          'Topic :: Utilities',
        ]
    )



import metacomm.combinatorics.all_pairs2
all_pairs = metacomm.combinatorics.all_pairs2.all_pairs2


"""
Demo of filtering capabilities
"""

# sample parameters are is taken from 
# http://www.stsc.hill.af.mil/consulting/sw_testing/improvement/cst.html


parameters = [ [ "Brand X", "Brand Y" ]
             , [ "98", "NT", "2000", "XP"]
             , [ "Internal", "Modem" ]
             , [ "Salaried", "Hourly", "Part-Time", "Contr." ]
             , [ 6, 10, 15, 30, 60 ]
             ]

def is_valid_combination( row ):
    """
    Should return True if combination is valid and False otherwise.
    
    Test row that is passed here can be incomplete.
    To prevent search for unnecessary items filtering function
    is executed with found subset of data to validate it.
    """

    n = len(row)
    if n>1:
        # Brand Y does not support Windows 98
        if "98" == row[1] and "Brand Y" == row[0]:
            return False
        # Brand X does not work with XP
        if "XP" == row[1] and "Brand X" == row[0]:
            return False
    if n > 4: 
        # Contractors are billed in 30 min increments
        if "Contr." == row[3] and row[4] < 30:
            return False
    
    return True
          
             
pairwise = all_pairs( parameters, filter_func = is_valid_combination )

for i, v in enumerate(pairwise):
    print "%i:\t%s" % (i, str(v))





        


import pairs_storage
from combinatorics import xuniqueCombinations 

class item:
    def __init__(self, id, value):
        self.id        = id
        self.value     = value
        self.weights = []
        
    def __str__(self):
        return str(self.__dict__)

    
def get_max_comb_number( arr, n ):
    items = [len(x) for x in arr]
    #print items
    f = lambda x,y:x*y
    total = sum([ reduce(f, z) for z in xuniqueCombinations( items, n) ])
    return total
    
    
class all_pairs2:
    def __iter__( self ):
        return self
        
    def __init__( self, options, filter_func = lambda x: True, previously_tested = [[]], n = 2 ):
        """
        TODO: check that input arrays are:
            - (optional) has no duplicated values inside single array / or compress such values
        """
        
        if len( options ) < 2:
            raise Exception("must provide more than one option")
        for arr in options:
            if not len(arr):
                raise Exception("option arrays must have at least one item")

        self.__filter_func = filter_func
        self.__n = n
        self.__pairs = pairs_storage.pairs_storage(n)
        self.__max_unique_pairs_expected = get_max_comb_number( options, n )
        self.__working_arr = []

        for i in range( len( options )):
            self.__working_arr.append( [ item("a%iv%i" % (i,j), value) \
                                         for j, value in enumerate(options[i] ) ] )

        for arr in previously_tested:
            if len(arr) == 0:
                continue
            elif len(arr) != len(self.__working_arr):
                raise Exception("previously tested combination is not complete")
            if not self.__filter_func(arr):
                raise Exception("invalid tested combination is provided")
            tested = []
            for i, val in enumerate(arr):
                idxs = [item(node.id, 0) for node in self.__working_arr[i] if node.value == val]
                if len(idxs) != 1:
                    raise Exception("value from previously tested combination is not found in the options or found more than once")
                tested.append(idxs[0])
            self.__pairs.add_sequence(tested)

    def next( self ):
        assert( len(self.__pairs) <= self.__max_unique_pairs_expected )
        p = self.__pairs
        if len(self.__pairs) == self.__max_unique_pairs_expected:
            # no reasons to search further - all pairs are found
            raise StopIteration
        
        previous_unique_pairs_count= len(self.__pairs)
        chosen_values_arr          = [None] * len(self.__working_arr)
        indexes                    = [None] * len(self.__working_arr)
        
        direction = 1
        i = 0
        
        while -1 < i < len(self.__working_arr):
            if direction == 1: # move forward
                self.resort_working_array( chosen_values_arr[:i], i )
                indexes[i] = 0
            elif direction == 0 or direction == -1: # scan current array or go back
                indexes[i] += 1
                if indexes[i] >= len( self.__working_arr[i] ):
                    direction = -1
                    if i == 0:
                        raise StopIteration
                    i += direction    
                    continue
                direction = 0
            else:
                raise Exception("next(): unknown 'direction' code.")
                    
            chosen_values_arr[i] =  self.__working_arr[i][ indexes[i] ]
            
            if self.__filter_func( self.get_values_array( chosen_values_arr[:i+1] ) ):
                assert(direction > -1)
                direction = 1
            else:
                direction = 0
            i += direction    
        
        if  len( self.__working_arr ) != len(chosen_values_arr):
            raise StopIteration
        
        self.__pairs.add_sequence( chosen_values_arr )

        if len(self.__pairs) == previous_unique_pairs_count:
            # could not find new unique pairs - stop
            raise StopIteration
        
        # replace returned array elements with real values and return it
        return self.get_values_array( chosen_values_arr )
        
    def get_values_array( self, arr ):
        return [ item.value for item in arr ]
    
    def resort_working_array( self, chosen_values_arr, num ):
        for item in self.__working_arr[num]:
            data_node = self.__pairs.get_node_info( item )
            
            new_combs = []
            for i in range(0, self.__n):
                # numbers of new combinations to be created if this item is appended to array
                new_combs.append( set([pairs_storage.key(z) for z in xuniqueCombinations( chosen_values_arr+[item], i+1)]) - self.__pairs.get_combs()[i] )
            # weighting the node
            item.weights =  [ -len(new_combs[-1]) ]    # node that creates most of new pairs is the best
            item.weights += [ len(data_node.out) ] # less used outbound connections most likely to produce more new pairs while search continues
            item.weights += [ len(x) for x in reversed(new_combs[:-1])]
            item.weights += [ -data_node.counter ]  # less used node is better
            item.weights += [ -len(data_node.in_) ] # otherwise we will prefer node with most of free inbound connections; somehow it works out better ;)
            
        self.__working_arr[num].sort( lambda a,b: cmp(a.weights, b.weights) )

    # statistics, internal stuff        
    def get_pairs_found( self ):
        return self.__pairs

__export__ = [ all_pairs2, get_max_comb_number ]



import metacomm.combinatorics.all_pairs2
all_pairs = metacomm.combinatorics.all_pairs2.all_pairs2

"""
Demo of the basic functionality - just getting pairwise/n-wise combinations
"""


# sample parameters are is taken from 
# http://www.stsc.hill.af.mil/consulting/sw_testing/improvement/cst.html


parameters = [ [ "Brand X", "Brand Y" ]
             , [ "98", "NT", "2000", "XP"]
             , [ "Internal", "Modem" ]
             , [ "Salaried", "Hourly", "Part-Time", "Contr." ]
             , [ 6, 10, 15, 30, 60 ]
             ]

triplewise = all_pairs( parameters, n=3 )

print "TRIPLEWISE:"
for i, v in enumerate(triplewise):
    print "%i:\t%s" % (i, str(v))
    
    